from __future__ import unicode_literals

import re

from mopidy.models import Ref

import pycountry


def unparse_uri(variant, identifier):
    return b'dirble:%s:%s' % (variant, identifier)


def parse_uri(uri):
    result = re.findall(r'^dirble:([a-z]+)(?::(\d+|[a-z]{2}))?$', uri)
    if result:
        return result[0]
    return None, None


def station_to_ref(station):
    name = station.get('name', station['streamurl']).strip()
    country = station.get('country', '??')
    uri = unparse_uri('station', station['id'])
    return Ref.track(uri=uri, name='%s - %s' % (country, name))


def category_to_ref(category):
    uri = unparse_uri('category', category['id'])
    return Ref.directory(uri=uri, name=category.get('name', uri))


def country_to_ref(country_code):
    uri = unparse_uri('country', country_code.lower())
    try:
        country = pycountry.countries.get(alpha2=country_code.upper())
        return Ref.directory(uri=uri, name=country.name)
    except KeyError:
        return Ref.directory(uri=uri, name=country_code.upper())



from __future__ import unicode_literals

import json
import logging
import time
import urllib2

logger = logging.getLogger(__name__)


class Dirble(object):
    """Light wrapper for Dirble API lookup.

    Important things to note:
    - The client will do exponential back off when requests fail or timeout.
    - The client will cache results aggressively.
    - Failed requests will return an empty default type appropriate for the
      lookup in question, normally a empty dict or list.
    - The data returned comes direct from the API's JSON.
    - For stations, only the minimal data set returned by station lists is
      guaranteed to be there.
    """
    def __init__(self, api_key, timeout):
        self._base_uri = 'http://api.dirble.com/v1/%s/apikey/' + api_key
        self._cache = {}
        self._stations = {}
        self._timeout = timeout / 1000.0
        self._backoff_until = time.time()
        self._backoff_max = 60
        self._backoff = 1

    def flush(self):
        self._cache = {}
        self._stations = {}

    def categories(self, category=None):
        if category:
            path = '/primaryid/%s' % category
            return self._fetch('childCategories', path, [])
        else:
            return self._fetch('primaryCategories', '', [])

    def stations(self, category=None, country=None):
        if category and not country:
            path = '/id/%s' % category
            stations = self._fetch('stations', path, [])
        elif country and not category:
            path = '/country/%s' % country.lower()
            stations = self._fetch('country', path, [])
        else:
            return []

        for station in stations:
            self._stations.setdefault(station['id'], station)
        return stations

    def station(self, identifier):
        identifier = int(identifier)  # Ensure we are consistent for cache key.
        if identifier in self._stations:
            return self._stations[identifier]
        path = '/id/%s' % identifier
        station = self._fetch('station', path, {})
        if station:
            self._stations.setdefault(station['id'], station)
        return station

    def _fetch(self, variant, path, default):
        uri = (self._base_uri % variant) + path
        if uri in self._cache:
            logger.debug('Cache hit: %s', uri)
            return self._cache[uri]

        if time.time() < self._backoff_until:
            logger.debug('Back off fallback used: %s', uri)
            return default

        logger.debug('Fetching: %s', uri)
        try:
            fp = urllib2.urlopen(uri, timeout=self._timeout)
            data = json.load(fp)
            self._cache[uri] = data
            self._backoff = 1
            return data
        except urllib2.HTTPError as e:
            logger.debug('Fetch failed, HTTP %s: %s', e.code, e.reason)
            if e.code == 404:
                self._cache[uri] = default
                return default
        except IOError as e:
            logger.debug('Fetch failed: %s', e)
        except ValueError as e:
            logger.warning('Fetch failed: %s', e)

        self._backoff = min(self._backoff_max, self._backoff*2)
        self._backoff_until = time.time() + self._backoff
        logger.debug('Entering back off mode for %d seconds.', self._backoff)
        return default



from __future__ import unicode_literals

import os

from mopidy import config, ext

__version__ = '0.1.1'


class Extension(ext.Extension):

    dist_name = 'Mopidy-Dirble'
    ext_name = 'dirble'
    version = __version__

    def get_default_config(self):
        conf_file = os.path.join(os.path.dirname(__file__), 'ext.conf')
        return config.read(conf_file)

    def get_config_schema(self):
        schema = super(Extension, self).get_config_schema()
        schema['api_key'] = config.String()
        schema['countries'] = config.List(optional=True)
        schema['timeout'] = config.Integer(minimum=0)
        return schema

    def setup(self, registry):
        from .actor import DirbleBackend
        registry.add('backend', DirbleBackend)



from __future__ import unicode_literals

import logging

import pykka

from mopidy import backend
from mopidy.models import Ref, Track

from . import client, translator

logger = logging.getLogger(__name__)


class DirbleBackend(pykka.ThreadingActor, backend.Backend):
    uri_schemes = ['dirble']

    def __init__(self, config, audio):
        super(DirbleBackend, self).__init__()
        self.dirble = client.Dirble(config['dirble']['api_key'],
                                    config['dirble']['timeout'])
        self.countries = config['dirble']['countries']
        self.library = DirbleLibrary(backend=self)
        self.playback = DirblePlayback(audio=audio, backend=self)


class DirbleLibrary(backend.LibraryProvider):
    root_directory = Ref.directory(uri='dirble:root', name='Dirble')

    # TODO: add countries when there is a lookup for countries with stations
    def browse(self, uri):
        result = []
        variant, identifier = translator.parse_uri(uri)

        if variant == 'root':
            for category in self.backend.dirble.categories():
                result.append(translator.category_to_ref(category))
            for country in self.backend.countries:
                result.append(translator.country_to_ref(country))
        elif variant == 'category' and identifier:
            for category in self.backend.dirble.categories(identifier):
                result.append(translator.category_to_ref(category))
            for station in self.backend.dirble.stations(category=identifier):
                result.append(translator.station_to_ref(station))
        elif variant == 'country' and identifier:
            for station in self.backend.dirble.stations(country=identifier):
                result.append(translator.station_to_ref(station))
        else:
            logger.debug('Unknown URI: %s', uri)

        result.sort(key=lambda ref: ref.name)
        return result

    def refresh(self, uri=None):
        self.backend.dirble.flush()

    def lookup(self, uri):
        variant, identifier = translator.parse_uri(uri)
        if variant != 'station':
            return []
        station = self.backend.dirble.station(identifier)
        if not station:
            return []
        ref = translator.station_to_ref(station)
        return [Track(uri=ref.uri, name=ref.name)]

    def find_exact(self, query=None, uris=None):
        return None

    def search(self, query=None, uris=None):
        return None


class DirblePlayback(backend.PlaybackProvider):
    def change_track(self, track):
        variant, identifier = translator.parse_uri(track.uri)
        if variant != 'station':
            return False
        station = self.backend.dirble.station(identifier)
        track = track.copy(uri=station['streamurl'])
        return super(DirblePlayback, self).change_track(track)



from __future__ import unicode_literals

import unittest

from mopidy_dirble import Extension


class ExtensionTest(unittest.TestCase):

    def test_get_default_config(self):
        ext = Extension()

        config = ext.get_default_config()

        self.assertIn('[dirble]', config)
        self.assertIn('enabled = true', config)

    def test_get_config_schema(self):
        ext = Extension()

        schema = ext.get_config_schema()
        self.assertIn('api_key', schema)

    # TODO Write more tests
